@@ -9,7 +9,8 @@
 static void
 test_launch_for_app_info (GAppInfo *appinfo)
 {
-  GError *error;
+  GError *error = NULL;
+  gboolean success;
   GFile *file;
   GList *l;
   const gchar *path;
@@ -21,30 +22,34 @@ test_launch_for_app_info (GAppInfo *appinfo)
       return;
     }
 
-  error = NULL;
-  g_assert (g_app_info_launch (appinfo, NULL, NULL, &error));
+  success = g_app_info_launch (appinfo, NULL, NULL, &error);
   g_assert_no_error (error);
+  g_assert_true (success);
 
-  g_assert (g_app_info_launch_uris (appinfo, NULL, NULL, &error));
+  success = g_app_info_launch_uris (appinfo, NULL, NULL, &error);
   g_assert_no_error (error);
+  g_assert_true (success);
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test.desktop", NULL);
   file = g_file_new_for_path (path);
   l = NULL;
   l = g_list_append (l, file);
 
-  g_assert (g_app_info_launch (appinfo, l, NULL, &error));
+  success = g_app_info_launch (appinfo, l, NULL, &error);
   g_assert_no_error (error);
+  g_assert_true (success);
   g_list_free (l);
   g_object_unref (file);
 
   l = NULL;
-  uri = g_strconcat ("file://", g_test_get_dir (G_TEST_DIST), "/appinfo-test.desktop", NULL);
+  uri = g_strconcat ("file://", g_test_get_dir (G_TEST_BUILT), "/appinfo-test.desktop", NULL);
   l = g_list_append (l, uri);
   l = g_list_append (l, "file:///etc/group#adm");
 
-  g_assert (g_app_info_launch_uris (appinfo, l, NULL, &error));
+  success = g_app_info_launch_uris (appinfo, l, NULL, &error);
   g_assert_no_error (error);
+  g_assert_true (success);
+
   g_list_free (l);
   g_free (uri);
 }
@@ -55,9 +60,14 @@ test_launch (void)
   GAppInfo *appinfo;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
-  g_assert (appinfo != NULL);
+
+  if (appinfo == NULL)
+    {
+      g_test_skip ("appinfo-test binary not installed");
+      return;
+    }
 
   test_launch_for_app_info (appinfo);
   g_object_unref (appinfo);
@@ -84,44 +94,50 @@ test_launch_no_app_id (void)
     "Keywords=keyword1;test keyword;\n"
     "Categories=GNOME;GTK;\n";
 
-  const char *exec_line_variants[] = {
-    "Exec=./appinfo-test --option %U %i --name %c --filename %k %m %%",
-    "Exec=./appinfo-test --option %u %i --name %c --filename %k %m %%"
-  };
-
+  gchar *exec_line_variants[2];
   gsize i;
 
+  exec_line_variants[0] = g_strdup_printf (
+      "Exec=%s/appinfo-test --option %%U %%i --name %%c --filename %%k %%m %%%%",
+      g_test_get_dir (G_TEST_BUILT));
+  exec_line_variants[1] = g_strdup_printf (
+      "Exec=%s/appinfo-test --option %%u %%i --name %%c --filename %%k %%m %%%%",
+      g_test_get_dir (G_TEST_BUILT));
+
   g_test_bug ("791337");
 
   for (i = 0; i < G_N_ELEMENTS (exec_line_variants); i++)
-  {
-    gchar *desktop_file_contents;
-    GKeyFile *fake_desktop_file;
-    GAppInfo *appinfo;
-    gboolean loaded;
+    {
+      gchar *desktop_file_contents;
+      GKeyFile *fake_desktop_file;
+      GAppInfo *appinfo;
+      gboolean loaded;
 
-    g_test_message ("Exec line variant #%" G_GSIZE_FORMAT, i);
+      g_test_message ("Exec line variant #%" G_GSIZE_FORMAT, i);
 
-    desktop_file_contents = g_strdup_printf ("%s\n%s",
-                                             desktop_file_base_contents,
-                                             exec_line_variants[i]);
+      desktop_file_contents = g_strdup_printf ("%s\n%s",
+                                               desktop_file_base_contents,
+                                               exec_line_variants[i]);
 
-    /* We load a desktop file from memory to force the app not
-     * to have an app ID, which would check different codepaths.
-     */
-    fake_desktop_file = g_key_file_new ();
-    loaded = g_key_file_load_from_data (fake_desktop_file, desktop_file_contents, -1, G_KEY_FILE_NONE, NULL);
-    g_assert_true (loaded);
+      /* We load a desktop file from memory to force the app not
+       * to have an app ID, which would check different codepaths.
+       */
+      fake_desktop_file = g_key_file_new ();
+      loaded = g_key_file_load_from_data (fake_desktop_file, desktop_file_contents, -1, G_KEY_FILE_NONE, NULL);
+      g_assert_true (loaded);
 
-    appinfo = (GAppInfo*)g_desktop_app_info_new_from_keyfile (fake_desktop_file);
-    g_assert (appinfo != NULL);
+      appinfo = (GAppInfo*)g_desktop_app_info_new_from_keyfile (fake_desktop_file);
+      g_assert_nonnull (appinfo);
 
-    test_launch_for_app_info (appinfo);
+      test_launch_for_app_info (appinfo);
 
-    g_free (desktop_file_contents);
-    g_object_unref (appinfo);
-    g_key_file_unref (fake_desktop_file);
-  }
+      g_free (desktop_file_contents);
+      g_object_unref (appinfo);
+      g_key_file_unref (fake_desktop_file);
+    }
+
+  g_free (exec_line_variants[1]);
+  g_free (exec_line_variants[0]);
 }
 
 static void
@@ -135,7 +151,7 @@ test_locale (const char *locale)
   g_setenv ("LANGUAGE", locale, TRUE);
   setlocale (LC_ALL, "");
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-static.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
 
   if (g_strcmp0 (locale, "C") == 0)
@@ -181,16 +197,17 @@ test_basic (void)
   GIcon *icon, *icon2;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-static.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
+  g_assert_nonnull (appinfo);
 
-  g_assert_cmpstr (g_app_info_get_id (appinfo), ==, "appinfo-test.desktop");
-  g_assert (strstr (g_app_info_get_executable (appinfo), "appinfo-test") != NULL);
+  g_assert_cmpstr (g_app_info_get_id (appinfo), ==, "appinfo-test-static.desktop");
+  g_assert_nonnull (strstr (g_app_info_get_executable (appinfo), "true"));
 
   icon = g_app_info_get_icon (appinfo);
-  g_assert (G_IS_THEMED_ICON (icon));
+  g_assert_true (G_IS_THEMED_ICON (icon));
   icon2 = g_themed_icon_new ("testicon");
-  g_assert (g_icon_equal (icon, icon2));
+  g_assert_true (g_icon_equal (icon, icon2));
   g_object_unref (icon2);
 
   appinfo2 = g_app_info_dup (appinfo);
@@ -207,19 +224,26 @@ test_show_in (void)
   GAppInfo *appinfo;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
-  g_assert (g_app_info_should_show (appinfo));
+
+  if (appinfo == NULL)
+    {
+      g_test_skip ("appinfo-test binary not installed");
+      return;
+    }
+
+  g_assert_true (g_app_info_should_show (appinfo));
   g_object_unref (appinfo);
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-gnome.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test-gnome.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
-  g_assert (g_app_info_should_show (appinfo));
+  g_assert_true (g_app_info_should_show (appinfo));
   g_object_unref (appinfo);
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-notgnome.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test-notgnome.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
-  g_assert (!g_app_info_should_show (appinfo));
+  g_assert_false (g_app_info_should_show (appinfo));
   g_object_unref (appinfo);
 }
 
@@ -239,12 +263,12 @@ test_commandline (void)
                                                 "cmdline-app-test",
                                                 G_APP_INFO_CREATE_SUPPORTS_URIS,
                                                 &error);
-  g_assert (appinfo != NULL);
   g_assert_no_error (error);
+  g_assert_nonnull (appinfo);
   g_assert_cmpstr (g_app_info_get_name (appinfo), ==, "cmdline-app-test");
   g_assert_cmpstr (g_app_info_get_commandline (appinfo), ==, cmdline_out);
-  g_assert (g_app_info_supports_uris (appinfo));
-  g_assert (!g_app_info_supports_files (appinfo));
+  g_assert_true (g_app_info_supports_uris (appinfo));
+  g_assert_false (g_app_info_supports_files (appinfo));
 
   g_object_unref (appinfo);
 
@@ -256,12 +280,12 @@ test_commandline (void)
                                                 "cmdline-app-test",
                                                 G_APP_INFO_CREATE_NONE,
                                                 &error);
-  g_assert (appinfo != NULL);
   g_assert_no_error (error);
+  g_assert_nonnull (appinfo);
   g_assert_cmpstr (g_app_info_get_name (appinfo), ==, "cmdline-app-test");
   g_assert_cmpstr (g_app_info_get_commandline (appinfo), ==, cmdline_out);
-  g_assert (!g_app_info_supports_uris (appinfo));
-  g_assert (g_app_info_supports_files (appinfo));
+  g_assert_false (g_app_info_supports_uris (appinfo));
+  g_assert_true (g_app_info_supports_files (appinfo));
 
   g_object_unref (appinfo);
 
@@ -286,10 +310,10 @@ test_launch_context (void)
                                                 NULL);
 
   str = g_app_launch_context_get_display (context, appinfo, NULL);
-  g_assert (str == NULL);
+  g_assert_null (str);
 
   str = g_app_launch_context_get_startup_notify_id (context, appinfo, NULL);
-  g_assert (str == NULL);
+  g_assert_null (str);
 
   g_object_unref (appinfo);
   g_object_unref (context);
@@ -308,8 +332,8 @@ launched (GAppLaunchContext *context,
   gint pid;
 
   pid = 0;
-  g_assert (g_variant_lookup (platform_data, "pid", "i", &pid));
-  g_assert (pid != 0);
+  g_assert_true (g_variant_lookup (platform_data, "pid", "i", &pid));
+  g_assert_cmpint (pid, !=, 0);
 
   launched_reached = TRUE;
 }
@@ -327,6 +351,7 @@ test_launch_context_signals (void)
   GAppLaunchContext *context;
   GAppInfo *appinfo;
   GError *error = NULL;
+  gboolean success;
   gchar *cmdline;
 
   cmdline = g_strconcat (g_test_get_dir (G_TEST_BUILT), "/appinfo-test --option", NULL);
@@ -339,11 +364,11 @@ test_launch_context_signals (void)
                                                 G_APP_INFO_CREATE_SUPPORTS_URIS,
                                                 NULL);
 
-  error = NULL;
-  g_assert (g_app_info_launch (appinfo, NULL, context, &error));
+  success = g_app_info_launch (appinfo, NULL, context, &error);
   g_assert_no_error (error);
+  g_assert_true (success);
 
-  g_assert (launched_reached);
+  g_assert_true (launched_reached);
 
   g_object_unref (appinfo);
   g_object_unref (context);
@@ -357,10 +382,10 @@ test_tryexec (void)
   GAppInfo *appinfo;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test2.desktop", NULL);
+  path = g_test_get_filename (G_TEST_BUILT, "appinfo-test2.desktop", NULL);
   appinfo = (GAppInfo*)g_desktop_app_info_new_from_filename (path);
 
-  g_assert (appinfo == NULL);
+  g_assert_null (appinfo);
 }
 
 /* Test that we can set an appinfo as default for a mime type or
@@ -371,7 +396,7 @@ test_associations (void)
 {
   GAppInfo *appinfo;
   GAppInfo *appinfo2;
-  GError *error;
+  GError *error = NULL;
   gboolean result;
   GList *list;
   gchar *cmdline;
@@ -383,33 +408,31 @@ test_associations (void)
                                                 NULL);
   g_free (cmdline);
 
-  error = NULL;
   result = g_app_info_set_as_default_for_type (appinfo, "application/x-glib-test", &error);
-
-  g_assert (result);
   g_assert_no_error (error);
+  g_assert_true (result);
 
   appinfo2 = g_app_info_get_default_for_type ("application/x-glib-test", FALSE);
 
-  g_assert (appinfo2);
+  g_assert_nonnull (appinfo2);
   g_assert_cmpstr (g_app_info_get_commandline (appinfo), ==, g_app_info_get_commandline (appinfo2));
 
   g_object_unref (appinfo2);
 
   result = g_app_info_set_as_default_for_extension (appinfo, "gio-tests", &error);
-  g_assert (result);
   g_assert_no_error (error);
+  g_assert_true (result);
 
   appinfo2 = g_app_info_get_default_for_type ("application/x-extension-gio-tests", FALSE);
 
-  g_assert (appinfo2);
+  g_assert_nonnull (appinfo2);
   g_assert_cmpstr (g_app_info_get_commandline (appinfo), ==, g_app_info_get_commandline (appinfo2));
 
   g_object_unref (appinfo2);
 
   result = g_app_info_add_supports_type (appinfo, "application/x-gio-test", &error);
-  g_assert (result);
   g_assert_no_error (error);
+  g_assert_true (result);
 
   list = g_app_info_get_all_for_type ("application/x-gio-test");
   g_assert_cmpint (g_list_length (list), ==, 1);
@@ -418,12 +441,13 @@ test_associations (void)
   g_object_unref (appinfo2);
   g_list_free (list);
 
-  g_assert (g_app_info_can_remove_supports_type (appinfo));
-  g_assert (g_app_info_remove_supports_type (appinfo, "application/x-gio-test", &error));
+  g_assert_true (g_app_info_can_remove_supports_type (appinfo));
+  result = g_app_info_remove_supports_type (appinfo, "application/x-gio-test", &error);
   g_assert_no_error (error);
+  g_assert_true (result);
 
-  g_assert (g_app_info_can_delete (appinfo));
-  g_assert (g_app_info_delete (appinfo));
+  g_assert_true (g_app_info_can_delete (appinfo));
+  g_assert_true (g_app_info_delete (appinfo));
   g_object_unref (appinfo);
 }
 
@@ -442,8 +466,8 @@ test_environment (void)
 
   env = g_app_launch_context_get_environment (ctx);
 
-  g_assert (g_environ_getenv (env, "FOO") == NULL);
-  g_assert (g_environ_getenv (env, "BLA") == NULL);
+  g_assert_null (g_environ_getenv (env, "FOO"));
+  g_assert_null (g_environ_getenv (env, "BLA"));
   g_assert_cmpstr (g_environ_getenv (env, "PATH"), ==, path);
 
   g_strfreev (env);
@@ -465,7 +489,7 @@ test_environment (void)
   env = g_app_launch_context_get_environment (ctx);
 
   g_assert_cmpstr (g_environ_getenv (env, "FOO"), ==, "baz");
-  g_assert (g_environ_getenv (env, "BLA") == NULL);
+  g_assert_null (g_environ_getenv (env, "BLA"));
 
   g_strfreev (env);
 
@@ -479,7 +503,7 @@ test_startup_wm_class (void)
   const char *wm_class;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-static.desktop", NULL);
   appinfo = g_desktop_app_info_new_from_filename (path);
   wm_class = g_desktop_app_info_get_startup_wm_class (appinfo);
 
@@ -495,7 +519,7 @@ test_supported_types (void)
   const char * const *content_types;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-static.desktop", NULL);
   appinfo = G_APP_INFO (g_desktop_app_info_new_from_filename (path));
   content_types = g_app_info_get_supported_types (appinfo);
 
@@ -512,53 +536,53 @@ test_from_keyfile (void)
   GKeyFile *kf;
   GError *error = NULL;
   const gchar *categories;
+  gchar **categories_list;
+  gsize categories_count;
   gchar **keywords;
   const gchar *file;
   const gchar *name;
   const gchar *path;
 
-  path = g_test_get_filename (G_TEST_DIST, "appinfo-test.desktop", NULL);
+  path = g_test_get_filename (G_TEST_DIST, "appinfo-test-static.desktop", NULL);
   kf = g_key_file_new ();
   g_key_file_load_from_file (kf, path, G_KEY_FILE_NONE, &error);
   g_assert_no_error (error);
   info = g_desktop_app_info_new_from_keyfile (kf);
-  g_key_file_free (kf);
-  g_assert (info != NULL);
+  g_key_file_unref (kf);
+  g_assert_nonnull (info);
 
   g_object_get (info, "filename", &file, NULL);
-  g_assert (file == NULL);
+  g_assert_null (file);
 
   file = g_desktop_app_info_get_filename (info);
-  g_assert (file == NULL);
+  g_assert_null (file);
   categories = g_desktop_app_info_get_categories (info);
   g_assert_cmpstr (categories, ==, "GNOME;GTK;");
+  categories_list = g_desktop_app_info_get_string_list (info, "Categories", &categories_count);
+  g_assert_cmpint (categories_count, ==, 2);
+  g_assert_cmpint (g_strv_length (categories_list), ==, 2);
+  g_assert_cmpstr (categories_list[0], ==, "GNOME");
+  g_assert_cmpstr (categories_list[1], ==, "GTK");
   keywords = (gchar **)g_desktop_app_info_get_keywords (info);
   g_assert_cmpint (g_strv_length (keywords), ==, 2);
   g_assert_cmpstr (keywords[0], ==, "keyword1");
   g_assert_cmpstr (keywords[1], ==, "test keyword");
   name = g_desktop_app_info_get_generic_name (info);
   g_assert_cmpstr (name, ==, "generic-appinfo-test");
-  g_assert (!g_desktop_app_info_get_nodisplay (info));
+  g_assert_false (g_desktop_app_info_get_nodisplay (info));
 
+  g_strfreev (categories_list);
   g_object_unref (info);
 }
 
 int
 main (int argc, char *argv[])
 {
-  const gchar *build_dir;
-
   g_setenv ("XDG_CURRENT_DESKTOP", "GNOME", TRUE);
 
-  g_test_init (&argc, &argv, NULL);
+  g_test_init (&argc, &argv, G_TEST_OPTION_ISOLATE_DIRS, NULL);
   g_test_bug_base ("https://bugzilla.gnome.org/show_bug.cgi?id=");
 
-  /* With Meson build we need to change into right directory, so that the
-   * appinfo-test binary can be found. */
-  build_dir = g_getenv ("G_TEST_BUILDDIR");
-  if (build_dir)
-    g_chdir (build_dir);
-
   g_test_add_func ("/appinfo/basic", test_basic);
   g_test_add_func ("/appinfo/text", test_text);
   g_test_add_func ("/appinfo/launch", test_launch);
